import pytest
from omnisci.mapd.ttypes import TColumnType
from omnisci.common.ttypes import TTypeInfo
from pymapd import OperationalError, connect
from pymapd.connection import _parse_uri, ConnectionInfo
from pymapd.exceptions import Error
from pymapd._parsers import ColumnDetails, _extract_column_details


@pytest.mark.usefixtures("mapd_server")
class TestConnect:

    def test_host_specified(self):
        with pytest.raises(TypeError):
            connect(user='foo')

    def test_raises_right_exception(self):
        with pytest.raises(OperationalError):
            connect(host='localhost', protocol='binary', port=1234)

    def test_close(self):
        conn = connect(user='admin', password='HyperInteractive',
                       host='localhost', dbname='omnisci')
        assert conn.closed == 0
        conn.close()
        assert conn.closed == 1

    def test_commit_noop(self, con):
        result = con.commit()  # it worked
        assert result is None

    def test_bad_protocol(self, mock_client):
        with pytest.raises(ValueError) as m:
            connect(user='user', host='localhost', dbname='dbname',
                    protocol='fake-proto')
        assert m.match('fake-proto')

    def test_session_logon_success(self):
        conn = connect(user='admin', password='HyperInteractive',
                       host='localhost', dbname='omnisci')
        sessionid = conn._session
        connnew = connect(sessionid=sessionid, host='localhost')
        assert connnew._session == sessionid

    def test_session_logon_failure(self):
        sessionid = 'ILoveDancingOnTables'
        with pytest.raises(Error):
            connect(sessionid=sessionid, host='localhost', protocol='binary')

    def test_bad_binary_encryption_params(self):
        with pytest.raises(TypeError):
            connect(user='admin', host='localhost', dbname='omnisci',
                    protocol='http', validate=False)


class TestURI:

    def test_parse_uri(self):
        uri = ('omnisci://admin:HyperInteractive@localhost:6274/omnisci?'
               'protocol=binary')
        result = _parse_uri(uri)
        expected = ConnectionInfo("admin", "HyperInteractive", "localhost",
                                  6274, "omnisci", "binary", None, None)
        assert result == expected

    def test_both_raises(self):
        uri = ('omnisci://admin:HyperInteractive@localhost:6274/omnisci?'
               'protocol=binary')
        with pytest.raises(TypeError):
            connect(uri=uri, user='my user')


class TestExtras:
    def test_extract_row_details(self):
        data = [
            TColumnType(col_name='date_',
                        col_type=TTypeInfo(type=6, encoding=4, nullable=True,
                                           is_array=False, precision=0,
                                           scale=0, comp_param=32),
                        is_reserved_keyword=False, src_name=''),
            TColumnType(col_name='trans',
                        col_type=TTypeInfo(type=6, encoding=4, nullable=True,
                                           is_array=False, precision=0,
                                           scale=0, comp_param=32),
                        is_reserved_keyword=False, src_name=''),
            TColumnType(col_name='symbol',
                        col_type=TTypeInfo(type=6, encoding=4, nullable=True,
                                           is_array=False, precision=0,
                                           scale=0, comp_param=32),
                        is_reserved_keyword=False, src_name=''),
            TColumnType(col_name='qty',
                        col_type=TTypeInfo(type=1, encoding=0, nullable=True,
                                           is_array=False, precision=0,
                                           scale=0, comp_param=0),
                        is_reserved_keyword=False, src_name=''),
            TColumnType(col_name='price',
                        col_type=TTypeInfo(type=3, encoding=0, nullable=True,
                                           is_array=False, precision=0,
                                           scale=0, comp_param=0),
                        is_reserved_keyword=False, src_name=''),
            TColumnType(col_name='vol',
                        col_type=TTypeInfo(type=3, encoding=0, nullable=True,
                                           is_array=False, precision=0,
                                           scale=0, comp_param=0),
                        is_reserved_keyword=False, src_name='')]
        result = _extract_column_details(data)

        expected = [
            ColumnDetails(name='date_', type='STR', nullable=True, precision=0,
                          scale=0, comp_param=32, encoding='DICT',
                          is_array=False),
            ColumnDetails(name='trans', type='STR', nullable=True, precision=0,
                          scale=0, comp_param=32, encoding='DICT',
                          is_array=False),
            ColumnDetails(name='symbol', type='STR', nullable=True,
                          precision=0, scale=0, comp_param=32,
                          encoding='DICT',
                          is_array=False),
            ColumnDetails(name='qty', type='INT', nullable=True, precision=0,
                          scale=0, comp_param=0, encoding='NONE',
                          is_array=False),
            ColumnDetails(name='price', type='FLOAT', nullable=True,
                          precision=0, scale=0, comp_param=0, encoding='NONE',
                          is_array=False),
            ColumnDetails(name='vol', type='FLOAT', nullable=True, precision=0,
                          scale=0, comp_param=0, encoding='NONE',
                          is_array=False)
        ]
        assert result == expected



import pytest
pytest.importorskip('rbc')


def catch_udf_support_disabled(mth):
    def new_mth(self, con):
        try:
            return mth(self, con)
        except Exception as msg:
            if (type(msg).__name__ == 'TMapDException'
                and msg.error_msg.startswith(
                    'Runtime UDF registration is disabled')):
                print('Ignoring `%s` failure' % (msg.error_msg))
                return
            raise
    new_mth.__name__ = mth.__name__
    return new_mth


@pytest.mark.usefixtures("mapd_server")
class TestRuntimeUDF:

    @catch_udf_support_disabled
    def test_udf_incr(self, con):

        @con('int32(int32)', 'double(double)')
        def incr(x):
            return x + 1

        con.execute('drop table if exists test_udf_incr')
        con.execute('create table test_udf_incr (i4 integer, f8 double)')
        con.execute('insert into test_udf_incr values (1, 2.3);')
        con.execute('insert into test_udf_incr values (2, 3.4);')

        result = list(con.execute('select i4, incr(i4) from test_udf_incr'))
        expected = [(1, 2), (2, 3)]
        assert result == expected

        result = list(con.execute('select f8, incr(f8) from test_udf_incr'))
        expected = [(2.3, 3.3), (3.4, 4.4)]
        assert result == expected

        con.execute('drop table if exists test_udf_incr')



import json
from base64 import b64decode, b64encode


def change_dashboard_sources(dashboard, remap):
    """
    Remap a dashboard to use a new table

    Parameters
    ----------
    dashboard: A dictionary containing the old dashboard state
    remap: A dictionary containing the new dashboard state to be mapped

    Returns
    -------
    dashboard: A base64 encoded json object containing the new dashboard state
    """
    dm = json.loads(dashboard.dashboard_metadata)
    tlst = map(str.strip, dm.get('table', '').split(','))
    tlst = [remap[t]['name'] if remap.get(t, {}).get('name',
                                                     {}) else t for t in tlst]
    dm['table'] = ', '.join(tlst)

    # Load our dashboard state into a python dictionary
    ds = json.loads(b64decode(dashboard.dashboard_state).decode())

    # Remap items in our old dashboard state to the new table name
    for old_table, defs in remap.items():
        new_table = defs.get('name', {}) or old_table
        if ds['dashboard']['table'] == old_table:
            ds['dashboard']['table'] = new_table

        # Change the name of the dashboard or use the old one
        ds['dashboard']['title'] = defs.get('title',
                                            {}) or ds['dashboard']['title']

        # Remap our datasources keys
        for key, val in ds['dashboard']['dataSources'].items():
            for col in val['columnMetadata']:
                if col['table'] == old_table:
                    col['table'] = new_table

        # Remap our charts to the new table
        for c, val in ds['charts'].items():
            if val.get('dataSource', None):
                if ds['charts'][c]['dataSource'] == old_table:
                    ds['charts'][c]['dataSource'] = new_table

            # Remap Our Dimensions to the new table
            i = 0
            for dim in val.get('dimensions', []):
                if dim.get('table', {}) == old_table:
                    ds['charts'][c]['dimensions'][i]['table'] = new_table
                if dim.get('selector', {}).get('table') == old_table:
                    (ds['charts'][c]['dimensions'][i]
                        ['selector']['table']) = new_table
                i += 1

            # Remap Our Measures to the new table
            i = 0
            for m in val.get('measures', []):
                if m.get('table', None) == old_table:
                    ds['charts'][c]['measures'][i]['table'] = new_table
                i += 1

            # Remap Our Layers to the new table
            il = 0
            for layer in val.get('layers', []):
                im = 0
                if layer.get('dataSource', {}) == old_table:
                    ds['charts'][c]['layers'][il]['dataSource'] = new_table
                for measure in layer.get('measures', []):
                    if measure.get('table', None) == old_table:
                        (ds
                            ['charts']
                            [c]
                            ['layers']
                            [il]
                            ['measures']
                            [im]
                            ['table']
                         ) = new_table
                    im += 1
                il += 1
        (ds
            ['dashboard']
            ['dataSources']
            [new_table]
         ) = ds['dashboard']['dataSources'].pop(old_table)

    # Convert our new dashboard state to a python object
    dashboard.dashboard_state = b64encode(json.dumps(ds).encode()).decode()
    dashboard.dashboard_metadata = json.dumps(dm)
    return dashboard



import pyarrow as pa
import pandas as pd
import numpy as np  # noqa
import pandas.util.testing as tm  # noqa
from pymapd._parsers import _load_schema, _load_data  # noqa


def make_data_batch():
    np.random.seed(1234)
    depdelay = np.random.randint(-5, 30, size=10, dtype=np.int16)
    arrdelay = np.random.randint(-15, 30, size=10, dtype=np.int16)
    depdelay_ = pa.array(depdelay)
    arrdelay_ = pa.array(arrdelay)
    batch = pa.RecordBatch.from_arrays([depdelay_, arrdelay_],
                                       ['depdelay', 'arrdelay'])
    return (depdelay, arrdelay), batch


(depdelay, arrdelay), batch = make_data_batch()
schema_data = batch.schema.serialize().to_pybytes()
data_data = batch.serialize().to_pybytes()


class TestIPC:
    def test_parse_schema(self):
        buf = pa.py_buffer(schema_data)
        result = _load_schema(buf)
        expected = pa.schema([
            pa.field("depdelay", pa.int16()),
            pa.field("arrdelay", pa.int16())
        ])
        assert result.equals(expected)

    def test_parse_data(self):
        buf = pa.py_buffer(data_data)
        schema = pa.schema([
            pa.field("depdelay", pa.int16()),
            pa.field("arrdelay", pa.int16())
        ])
        result = _load_data(buf, schema)
        expected = pd.DataFrame({
            "depdelay": depdelay,
            "arrdelay": arrdelay,
        })[['depdelay', 'arrdelay']]
        tm.assert_frame_equal(result, expected)



"""
Setter and Getter for TDataFrame
"""


def set_tdf(self, tdf):
    """Assigns a TDataFrame to cudf/pandas Dataframe

        Parameters
        ----------
        tdf: TDataFrame
            A SQL select statement

        Example
        -------
        >>> df.set_tdf(tdf)
        """
    self._tdf = tdf


def get_tdf(self):
    """Returns assigned TDataFrame.

        Example
        -------
        >>> df.get_tdf()
        TDataFrame(sm_handle=b'\xa30q%', sm_size=632, df_handle=b'\x90a>
        \x06\x00\x00\x00\x00\xe0\xe6\x00\x00\x00\x00\x00\x00\xe0|E\x00\x00
        \x00\xca\x01\xd0\xc1"\x03\x00\\', df_size=4553952)
        """

    return self._tdf



old_dashboard_name = "old_dashboard"
old_dashboard_state = {
    "charts": {
        "1": {
            "autoSize": True,
            "areFiltersInverse": False,
            "cap": 12,
            "renderArea": False,
            "color": {
                "type": "none"
            },
            "colorDomain": None,
            "dcFlag": 1,
            "densityAccumulatorEnabled": True,
            "dimensions": [
                {
                    "inactive": False,
                    "name": None,
                    "isRequired": False,
                    "isError": False,
                    "table": "test_data_no_nulls_ipc",
                    "type": "STR",
                    "precision": 0,
                    "is_array": False,
                    "is_dict": True,
                    "name_is_ambiguous": False,
                    "label": "text_",
                    "value": "text_",
                    "custom": False,
                    "axisLabel": None,
                    "max_val": None,
                    "min_val": None,
                    "maxBinSize": None,
                    "currentHighValue": None,
                    "currentLowValue": None,
                    "autobin": False,
                    "numOfBins": None,
                    "isBinned": False,
                    "isBinnable": False
                },
                {
                    "isRequired": False,
                    "isError": False
                }
            ],
            "elasticX": True,
            "elasticY": True,
            "filters": [],
            "geoJson": None,
            "loading": True,
            "measures": [
                {
                    "inactive": False,
                    "name": "col0",
                    "isRequired": False,
                    "isError": False,
                    "label": "# Records",
                    "value": "*",
                    "type": "SMALLINT",
                    "custom": False,
                    "axisLabel": None,
                    "minMax": None,
                    "categories": None,
                    "colorType": "quantitative",
                    "aggType": "Count",
                    "originIndex": 0
                },
                {
                    "isRequired": False,
                    "isError": False
                }
            ],
            "rangeChartEnabled": False,
            "rangeFilter": [],
            "savedColors": {},
            "sortColumn": None,
            "ticks": 3,
            "title": "",
            "type": "table",
            "showOther": False,
            "showNullDimensions": True,
            "markTypes": [],
            "multiSources": {},
            "hasError": False,
            "dataSource": "test_data_no_nulls_ipc",
            "hoverSelectedColumns": []
        },
        "test_data_no_nulls_ipc": {
            "dcFlag": 2,
            "loading": False,
            "color": {
                "defaultOtherDomain": "Default"
            }
        }
    },
    "ui": {
        "showFilterPanel": False,
        "showClearFiltersDropdown": False,
        "banner": {
            "open": False,
            "type": None
        },
        "modal": {
            "type": None,
            "open": False,
            "content": "",
            "header": "",
            "primaryAction": {
                "text": "OK"
            },
            "secondaryAction": {
                "text": "CANCEL"
            }
        },
        "selectorPillHover": {
            "shouldShowPrompt": False,
            "message": "",
            "top": 0
        },
        "selectorPositions": {
            "dimensions": [
                202.60000610351562,
                242.60000610351562,
                202.60000610351562,
                242.60000610351562,
                202.60000610351562
            ],
            "measures": [
                331.3999938964844,
                371.3999938964844,
                331.3999938964844,
                291.3999938964844
            ]
        },
        "nagScreenIsEnabled": False
    },
    "filters": [],
    "dashboard": {
        "id": None,
        "title": "old_dashboard_2",
        "chartContainers": [
            {
                "id": "1"
            }
        ],
        "table": None,
        "filtersId": [],
        "layout": [
            {
                "w": 10,
                "h": 10,
                "x": 0,
                "y": 0,
                "i": "1",
                "minW": 5,
                "minH": 5,
                "moved": False,
                "static": False
            }
        ],
        "currentDataSource": "test_data_no_nulls_ipc",
        "dataSources": {
            "test_data_no_nulls_ipc": {
                "alias": "A",
                "columnMetadata": [
                    {
                        "table": "test_data_no_nulls_ipc",
                        "type": "TINYINT",
                        "precision": 0,
                        "is_array": False,
                        "is_dict": False,
                        "name_is_ambiguous": False,
                        "label": "tinyint_",
                        "value": "tinyint_"
                    },
                    {
                        "table": "test_data_no_nulls_ipc",
                        "type": "SMALLINT",
                        "precision": 0,
                        "is_array": False,
                        "is_dict": False,
                        "name_is_ambiguous": False,
                        "label": "smallint_",
                        "value": "smallint_"
                    },
                    {
                        "table": "test_data_no_nulls_ipc",
                        "type": "INT",
                        "precision": 0,
                        "is_array": False,
                        "is_dict": False,
                        "name_is_ambiguous": False,
                        "label": "int_",
                        "value": "int_"
                    },
                    {
                        "table": "test_data_no_nulls_ipc",
                        "type": "BIGINT",
                        "precision": 0,
                        "is_array": False,
                        "is_dict": False,
                        "name_is_ambiguous": False,
                        "label": "bigint_",
                        "value": "bigint_"
                    },
                    {
                        "table": "test_data_no_nulls_ipc",
                        "type": "FLOAT",
                        "precision": 0,
                        "is_array": False,
                        "is_dict": False,
                        "name_is_ambiguous": False,
                        "label": "float_",
                        "value": "float_"
                    },
                    {
                        "table": "test_data_no_nulls_ipc",
                        "type": "DOUBLE",
                        "precision": 0,
                        "is_array": False,
                        "is_dict": False,
                        "name_is_ambiguous": False,
                        "label": "double_",
                        "value": "double_"
                    },
                    {
                        "table": "test_data_no_nulls_ipc",
                        "type": "DATE",
                        "precision": 0,
                        "is_array": False,
                        "is_dict": False,
                        "name_is_ambiguous": False,
                        "label": "date_",
                        "value": "date_"
                    },
                    {
                        "table": "test_data_no_nulls_ipc",
                        "type": "TIMESTAMP",
                        "precision": 0,
                        "is_array": False,
                        "is_dict": False,
                        "name_is_ambiguous": False,
                        "label": "datetime_",
                        "value": "datetime_"
                    },
                    {
                        "table": "test_data_no_nulls_ipc",
                        "type": "TIME",
                        "precision": 0,
                        "is_array": False,
                        "is_dict": False,
                        "name_is_ambiguous": False,
                        "label": "time_",
                        "value": "time_"
                    },
                    {
                        "table": "test_data_no_nulls_ipc",
                        "type": "STR",
                        "precision": 0,
                        "is_array": False,
                        "is_dict": True,
                        "name_is_ambiguous": False,
                        "label": "text_",
                        "value": "text_"
                    }
                ]
            }
        },
        "privileges": {},
        "saveLinkState": {
            "error": False,
            "request": False,
            "saveLinkId": None
        },
        "loadState": {
            "error": False,
            "dataAccessError": False,
            "complete": True,
            "request": False,
            "loadLinkId": None
        },
        "saveState": {
            "request": False,
            "error": False,
            "isSaved": False
        },
        "copyState": {
            "error": False,
            "request": False
        },
        "streaming": {
            "interval": 0,
            "request": False
        },
        "version": "4.6.1-20190429-02ec2e206b"
    },
    "machineLearning": {
        "training": False,
        "queued": False,
        "error": None,
        "predictionCol": None,
        "featureCols": [],
        "conditions": [],
        "algorithm": "Linear Regression",
        "nModels": 3,
        "nAlphas": 4,
        "nLambdas": 100,
        "nFolds": 4,
        "tensorFlowInput1": 0,
        "tensorFlowInput2": 0,
        "deepLearningInput1": "foo",
        "tableWithPredictions": None,
        "trainingTable": None,
        "predictedColumn": None,
        "accuracyColumn": None,
        "modelName": "",
        "predicting": False,
        "models": [],
        "complete": False,
        "trainResults": None,
        "accuracy": None,
        "jobId": None,
        "progress": {
            "steps": None,
            "current": None
        }
    }
}


"""
Tests that rely on a server running
"""
import datetime
from unittest import mock
import pytest
from pymapd import connect, ProgrammingError, DatabaseError
from pymapd.cursor import Cursor
from pymapd._parsers import Description, ColumnDetails
from omnisci.mapd.ttypes import TMapDException
from .data import dashboard_metadata
import json
import base64
import pandas as pd
import numpy as np
import pyarrow as pa
from pandas.api.types import is_object_dtype, is_categorical_dtype
import pandas.util.testing as tm
import textwrap
from .conftest import no_gpu

# XXX: Make it hashable to silence warnings; see if this can be done upstream
# This isn't a huge deal, but our testing context mangers for asserting
# exceptions need hashability
TMapDException.__hash__ = lambda x: id(x)


@pytest.mark.usefixtures("mapd_server")
class TestIntegration:

    def test_connect_binary(self):
        con = connect(user="admin", password='HyperInteractive',
                      host='localhost', port=6274, protocol='binary',
                      dbname='omnisci')
        assert con is not None

    def test_connect_http(self):
        con = connect(user="admin", password='HyperInteractive',
                      host='localhost', port=6278, protocol='http',
                      dbname='omnisci')
        assert con is not None

    def test_connect_uri(self):
        uri = ('omnisci://admin:HyperInteractive@localhost:6274/omnisci?'
               'protocol=binary')
        con = connect(uri=uri)
        assert con._user == 'admin'
        assert con._password == 'HyperInteractive'
        assert con._host == 'localhost'
        assert con._port == 6274
        assert con._dbname == 'omnisci'
        assert con._protocol == 'binary'

    def test_connect_uri_and_others_raises(self):
        uri = ('omnisci://admin:HyperInteractive@localhost:6274/omnisci?'
               'protocol=binary')
        with pytest.raises(TypeError):
            connect(username='omnisci', uri=uri)

    def test_invalid_sql(self, con):
        with pytest.raises(ProgrammingError) as r:
            con.cursor().execute("this is invalid;")
        r.match("Exception: Parse failed:")

    def test_nonexistant_table(self, con):
        with pytest.raises(DatabaseError) as r:
            con.cursor().execute("select it from fake_table;")
        r.match("Table 'FAKE_TABLE' does not exist|Object 'fake_table' not")

    def test_connection_execute(self, con):
        result = con.execute("drop table if exists FOO;")
        result = con.execute("create table FOO (a int);")
        assert isinstance(result, Cursor)
        con.execute("drop table if exists FOO;")

    def test_select_sets_description(self, con):

        c = con.cursor()
        c.execute('drop table if exists stocks;')
        create = ('create table stocks (date_ text, trans text, symbol text, '
                  'qty int, price float, vol float);')
        c.execute(create)
        i1 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14,1.1);"  # noqa
        i2 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','GOOG',100,12.14,1.2);"  # noqa

        c.execute(i1)
        c.execute(i2)

        c.execute("select * from stocks")
        expected = [
            Description('date_', 6, None, None, None, None, True),
            Description('trans', 6, None, None, None, None, True),
            Description('symbol', 6, None, None, None, None, True),
            Description('qty', 1, None, None, None, None, True),
            Description('price', 3, None, None, None, None, True),
            Description('vol', 3, None, None, None, None, True),
        ]
        assert c.description == expected
        c.execute('drop table if exists stocks;')

    def test_select_parametrized(self, con):

        c = con.cursor()
        c.execute('drop table if exists stocks;')
        create = ('create table stocks (date_ text, trans text, symbol text, '
                  'qty int, price float, vol float);')
        c.execute(create)
        i1 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14,1.1);"  # noqa
        i2 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','GOOG',100,12.14,1.2);"  # noqa

        c.execute(i1)
        c.execute(i2)

        c.execute('select symbol, qty from stocks where symbol = :symbol',
                  {'symbol': 'GOOG'})
        result = list(c)
        expected = [('GOOG', 100), ]  # noqa
        assert result == expected
        c.execute('drop table if exists stocks;')

    def test_executemany_parametrized(self, con):

        c = con.cursor()
        c.execute('drop table if exists stocks;')
        create = ('create table stocks (date_ text, trans text, symbol text, '
                  'qty int, price float, vol float);')
        c.execute(create)
        i1 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14,1.1);"  # noqa
        i2 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','GOOG',100,12.14,1.2);"  # noqa

        c.execute(i1)
        c.execute(i2)

        parameters = [{'symbol': 'GOOG'}, {'symbol': "RHAT"}]
        expected = [[('GOOG', 100)], [('RHAT', 100)]]
        query = 'select symbol, qty from stocks where symbol = :symbol'
        c = con.cursor()
        result = c.executemany(query, parameters)
        assert result == expected
        c.execute('drop table if exists stocks;')

    def test_executemany_parametrized_insert(self, con):

        c = con.cursor()
        c.execute('drop table if exists stocks;')
        create = ('create table stocks (date_ text, trans text, symbol text, '
                  'qty int, price float, vol float);')
        c.execute(create)
        i1 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14,1.1);"  # noqa
        i2 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','GOOG',100,12.14,1.2);"  # noqa

        c.execute(i1)
        c.execute(i2)

        c = con.cursor()
        c.execute("drop table if exists stocks2;")
        # Create table
        c.execute('CREATE TABLE stocks2 (symbol text, qty int);')
        params = [{"symbol": "GOOG", "qty": 10},
                  {"symbol": "AAPL", "qty": 20}]
        query = "INSERT INTO stocks2 VALUES (:symbol, :qty);"
        result = c.executemany(query, params)
        assert result == [[], []]  # TODO: not sure if this is standard
        c.execute("drop table stocks2;")
        c.execute('drop table if exists stocks;')

    @pytest.mark.parametrize('query, parameters', [
        ('select qty, price from stocks', None),
        ('select qty, price from stocks where qty=:qty', {'qty': 100}),
    ])
    def test_select_ipc_parametrized(self, con, query, parameters):

        c = con.cursor()
        c.execute('drop table if exists stocks;')
        create = ('create table stocks (date_ text, trans text, symbol text, '
                  'qty int, price float, vol float);')
        c.execute(create)
        i1 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14,1.1);"  # noqa
        i2 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','GOOG',100,12.14,1.2);"  # noqa

        c.execute(i1)
        c.execute(i2)

        result = con.select_ipc(query, parameters=parameters)
        expected = pd.DataFrame({
            "qty": np.array([100, 100], dtype=np.int32),
            "price": np.array([35.13999938964844, 12.140000343322754],
                              dtype=np.float32)
        })[['qty', 'price']]
        tm.assert_frame_equal(result, expected)
        c.execute('drop table if exists stocks;')

    def test_select_ipc_first_n(self, con):

        c = con.cursor()
        c.execute('drop table if exists stocks;')
        create = ('create table stocks (date_ text, trans text, symbol text, '
                  'qty int, price float, vol float);')
        c.execute(create)
        i1 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14,1.1);"  # noqa
        i2 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','GOOG',100,12.14,1.2);"  # noqa

        c.execute(i1)
        c.execute(i2)

        result = con.select_ipc("select * from stocks", first_n=1)
        assert len(result) == 1
        c.execute('drop table if exists stocks;')

    @pytest.mark.parametrize('query, parameters', [
        ('select qty, price from stocks', None),
        ('select qty, price from stocks where qty=:qty', {'qty': 100}),
    ])
    @pytest.mark.skipif(no_gpu(), reason="No GPU available")
    def test_select_ipc_gpu(self, con, query, parameters):

        from cudf.dataframe import DataFrame

        c = con.cursor()
        c.execute('drop table if exists stocks;')
        create = ('create table stocks (date_ text, trans text, symbol text, '
                  'qty int, price float, vol float);')
        c.execute(create)
        i1 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14,1.1);"  # noqa
        i2 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','GOOG',100,12.14,1.2);"  # noqa

        c.execute(i1)
        c.execute(i2)

        result = con.select_ipc_gpu("select qty, price from stocks")
        assert isinstance(result, DataFrame)

        dtypes = dict(qty=np.int32, price=np.float32)
        expected = pd.DataFrame([[100, 35.14], [100, 12.14]],
                                columns=['qty', 'price']).astype(dtypes)

        result = result.to_pandas()[['qty', 'price']]  # column order
        pd.testing.assert_frame_equal(result, expected)
        c.execute('drop table if exists stocks;')

    @pytest.mark.skipif(no_gpu(), reason="No GPU available")
    def test_select_gpu_first_n(self, con):

        c = con.cursor()
        c.execute('drop table if exists stocks;')
        create = ('create table stocks (date_ text, trans text, symbol text, '
                  'qty int, price float, vol float);')
        c.execute(create)
        i1 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14,1.1);"  # noqa
        i2 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','GOOG',100,12.14,1.2);"  # noqa

        c.execute(i1)
        c.execute(i2)

        result = con.select_ipc_gpu("select * from stocks", first_n=1)
        assert len(result) == 1
        c.execute('drop table if exists stocks;')

    def test_fetchone(self, con):

        c = con.cursor()
        c.execute('drop table if exists stocks;')
        create = ('create table stocks (date_ text, trans text, symbol text, '
                  'qty int, price float, vol float);')
        c.execute(create)
        i1 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14,1.1);"  # noqa
        i2 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','GOOG',100,12.14,1.2);"  # noqa

        c.execute(i1)
        c.execute(i2)

        c.execute("select symbol, qty from stocks")
        result = c.fetchone()
        expected = ('RHAT', 100)
        assert result == expected
        c.execute('drop table if exists stocks;')

    def test_fetchmany(self, con):

        c = con.cursor()
        c.execute('drop table if exists stocks;')
        create = ('create table stocks (date_ text, trans text, symbol text, '
                  'qty int, price float, vol float);')
        c.execute(create)
        i1 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14,1.1);"  # noqa
        i2 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','GOOG',100,12.14,1.2);"  # noqa

        c.execute(i1)
        c.execute(i2)

        c.execute("select symbol, qty from stocks")
        result = c.fetchmany()
        expected = [('RHAT', 100)]
        assert result == expected

        c.execute("select symbol, qty from stocks")
        result = c.fetchmany(size=10)
        expected = [('RHAT', 100), ('GOOG', 100)]
        assert result == expected
        c.execute('drop table if exists stocks;')

    def test_select_dates(self, con):

        c = con.cursor()
        c.execute('drop table if exists dates;')
        c.execute('create table dates (date_ DATE, datetime_ TIMESTAMP, '
                  'time_ TIME);')
        i1 = ("INSERT INTO dates VALUES ('2006-01-05','2006-01-01T12:00:00',"
              "'12:00:00');")
        i2 = ("INSERT INTO dates VALUES ('1901-12-14','1901-12-13T20:45:53',"
              "'23:59:00');")
        c.execute(i1)
        c.execute(i2)

        result = list(c.execute("select * from dates"))
        expected = [
            (datetime.date(2006, 1, 5),
             datetime.datetime(2006, 1, 1, 12),
             datetime.time(12)),
            (datetime.date(1901, 12, 14),
             datetime.datetime(1901, 12, 13, 20, 45, 53),
             datetime.time(23, 59)),
        ]
        assert result == expected
        c.execute('drop table if exists dates;')

    def test_dashboard_duplication_remap(self, con):
        # This test relies on the test_data_no_nulls_ipc table
        # Setup our testing variables
        old_dashboard_state = dashboard_metadata.old_dashboard_state
        old_dashboard_name = dashboard_metadata.old_dashboard_name
        new_dashboard_name = "new_test"
        meta_data = {"table": "test_data_no_nulls_ipc", "version": "v2"}
        remap = {
                    "test_data_no_nulls_ipc": {
                        "name": new_dashboard_name,
                        "title": new_dashboard_name
                        }
                }
        dashboards = []

        # Create testing dashboard
        try:
            dashboard_id = con._client.create_dashboard(
                session=con._session,
                dashboard_name=old_dashboard_name,
                dashboard_state=(
                    base64.b64encode(json.dumps(old_dashboard_state).encode(
                        "utf-8"))),
                image_hash="",
                dashboard_metadata=json.dumps(meta_data),
            )
        except TMapDException:
            dashboards = con._client.get_dashboards(con._session)
            for dash in dashboards:
                if dash.dashboard_name == old_dashboard_name:
                    con._client.delete_dashboard(con._session,
                                                 dash.dashboard_id)
                    break
            dashboard_id = con._client.create_dashboard(
                session=con._session,
                dashboard_name=old_dashboard_name,
                dashboard_state=(
                    base64.b64encode(json.dumps(old_dashboard_state).encode(
                        "utf-8"))),
                image_hash="",
                dashboard_metadata=json.dumps(meta_data),
            )

        # Duplicate and remap our dashboard
        try:
            dashboard_id = con.duplicate_dashboard(
                    dashboard_id, new_dashboard_name, remap
                )
        except TMapDException:
            dashboards = con._client.get_dashboards(con._session)
            for dash in dashboards:
                if dash.dashboard_name == new_dashboard_name:
                    con._client.delete_dashboard(con._session,
                                                 dash.dashboard_id)
                    break
            dashboard_id = con.duplicate_dashboard(
                    dashboard_id, new_dashboard_name, remap
                )

        # Get our new dashboard from the database
        d = con._client.get_dashboard(
                                        session=con._session,
                                        dashboard_id=dashboard_id
                                    )
        remapped_dashboard = json.loads(base64.b64decode(
            d.dashboard_state).decode())

        # Assert that the table and title changed
        assert remapped_dashboard['dashboard']['title'] == new_dashboard_name

        # Ensure the datasources change
        for key, val in remapped_dashboard['dashboard']['dataSources'].items():
            for col in val['columnMetadata']:
                assert col['table'] == new_dashboard_name


class TestOptionalImports:

    def test_select_gpu(self, con):
        with mock.patch.dict("sys.modules",
                             {"cudf": None, "cudf.dataframe": None}):
            with pytest.raises(ImportError) as m:
                con.select_ipc_gpu("select * from foo;")
        assert m.match("The 'cudf' package is required")


class TestExtras:

    def test_get_tables(self, con):

        c = con.cursor()
        c.execute('drop table if exists stocks;')
        create = ('create table stocks (date_ text, trans text, symbol text, '
                  'qty int, price float, vol float);')
        c.execute(create)
        i1 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14,1.1);"  # noqa
        i2 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','GOOG',100,12.14,1.2);"  # noqa

        c.execute(i1)
        c.execute(i2)

        result = con.get_tables()
        assert isinstance(result, list)
        assert 'stocks' in result
        c.execute('drop table if exists stocks;')

    def test_get_table_details(self, con):

        c = con.cursor()
        c.execute('drop table if exists stocks;')
        create = ('create table stocks (date_ text, trans text, symbol text, '
                  'qty int, price float, vol float, '
                  'exchanges TEXT [] ENCODING DICT(32));')
        c.execute(create)
        i1 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14,1.1,{'NYSE', 'NASDAQ', 'AMEX'});"  # noqa
        i2 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','GOOG',100,12.14,1.2,{'NYSE', 'NASDAQ'});"  # noqa

        c.execute(i1)
        c.execute(i2)

        result = con.get_table_details('stocks')
        expected = [
            ColumnDetails(name='date_', type='STR', nullable=True, precision=0,
                          scale=0, comp_param=32, encoding='DICT',
                          is_array=False),
            ColumnDetails(name='trans', type='STR', nullable=True, precision=0,
                          scale=0, comp_param=32, encoding='DICT',
                          is_array=False),
            ColumnDetails(name='symbol', type='STR', nullable=True,
                          precision=0, scale=0, comp_param=32,
                          encoding='DICT', is_array=False),
            ColumnDetails(name='qty', type='INT', nullable=True, precision=0,
                          scale=0, comp_param=0, encoding='NONE',
                          is_array=False),
            ColumnDetails(name='price', type='FLOAT', nullable=True,
                          precision=0, scale=0, comp_param=0, encoding='NONE',
                          is_array=False),
            ColumnDetails(name='vol', type='FLOAT', nullable=True, precision=0,
                          scale=0, comp_param=0, encoding='NONE',
                          is_array=False),
            ColumnDetails(name='exchanges', type='STR', nullable=True,
                          precision=0, scale=0, comp_param=32, encoding='DICT',
                          is_array=True)
        ]
        assert result == expected
        c.execute('drop table if exists stocks;')


class TestLoaders:

    @staticmethod
    def check_empty_insert(result, expected):
        assert len(result) == 3
        assert expected[0][0] == result[0][0]
        assert expected[0][2] == result[0][2]
        assert abs(expected[0][1] - result[0][1]) < 1e-7  # floating point

    def test_load_empty_table(self, con):

        con.execute("drop table if exists baz;")
        con.execute("create table baz (a int, b float, c text);")

        data = [(1, 1.1, 'a'),
                (2, 2.2, '2'),
                (3, 3.3, '3')]
        con.load_table("baz", data)
        result = sorted(con.execute("select * from baz"))
        self.check_empty_insert(result, data)
        con.execute("drop table if exists baz;")

    def test_load_empty_table_pandas(self, con):

        con.execute("drop table if exists baz;")
        con.execute("create table baz (a int, b float, c text);")

        data = [(1, 1.1, 'a'),
                (2, 2.2, '2'),
                (3, 3.3, '3')]
        df = pd.DataFrame(data, columns=list('abc'))
        con.load_table("baz", df, method='columnar')
        result = sorted(con.execute("select * from baz"))
        self.check_empty_insert(result, data)
        con.execute("drop table if exists baz;")

    def test_load_empty_table_arrow(self, con):

        con.execute("drop table if exists baz;")
        con.execute("create table baz (a int, b float, c text);")

        data = [(1, 1.1, 'a'),
                (2, 2.2, '2'),
                (3, 3.3, '3')]

        df = pd.DataFrame(data, columns=list('abc')).astype({
            'a': 'int32',
            'b': 'float32'
        })

        table = pa.Table.from_pandas(df, preserve_index=False)
        con.load_table("baz", table, method='arrow')
        result = sorted(con.execute("select * from baz"))
        self.check_empty_insert(result, data)
        con.execute("drop table if exists baz;")

    def test_load_table_columnar(self, con):

        con.execute("drop table if exists baz;")
        con.execute("create table baz (a int, b float, c text);")

        df = pd.DataFrame({"a": [1, 2, 3],
                           "b": [1.1, 2.2, 3.3],
                           "c": ['a', '2', '3']}, columns=['a', 'b', 'c'])
        con.load_table_columnar("baz", df)
        con.execute("drop table if exists baz;")

    def test_load_infer(self, con):

        con.execute("drop table if exists baz;")
        con.execute("create table baz (a int, b float, c text);")

        data = pd.DataFrame(
            {'a': np.array([0, 1], dtype=np.int32),
             'b': np.array([1.1, 2.2], dtype=np.float32),
             'c': ['a', 'b']}
        )
        con.load_table("baz", data)
        con.execute("drop table if exists baz;")

    def test_load_infer_bad(self, con):

        con.execute("drop table if exists baz;")
        con.execute("create table baz (a int, b float, c text);")

        with pytest.raises(TypeError):
            con.load_table("baz", [], method='thing')

        con.execute("drop table if exists baz;")

    def test_infer_non_pandas(self, con):

        con.execute("drop table if exists baz;")
        con.execute("create table baz (a int, b float, c text);")

        with pytest.raises(TypeError):
            con.load_table("baz", [], method='columnar')

        con.execute("drop table if exists baz;")

    def test_load_columnar_pandas_all(self, con):

        c = con.cursor()
        c.execute('drop table if exists all_types;')
        create = textwrap.dedent('''\
        create table all_types (
            boolean_ BOOLEAN,
            smallint_ SMALLINT,
            int_ INT,
            bigint_ BIGINT,
            float_ FLOAT,
            double_ DOUBLE,
            varchar_ VARCHAR(40),
            text_ TEXT,
            time_ TIME,
            timestamp_ TIMESTAMP,
            date_ DATE
        );''')
        # skipping decimal for now
        c.execute(create)

        data = pd.DataFrame({
            "boolean_": [True, False, True, False],
            "smallint_": np.array([0, 1, 0, 1], dtype=np.int16),
            "int_": np.array([0, 1, 0, 1], dtype=np.int32),
            "bigint_": np.array([0, 1, 0, 1], dtype=np.int64),
            "float_": np.array([0, 1, 0, 1], dtype=np.float32),
            "double_": np.array([0, 1, 0, 1], dtype=np.float64),
            "varchar_": ["a", "b", "a", "b"],
            "text_": ['a', 'b', 'a', 'b'],
            "time_": [datetime.time(0, 11, 59), datetime.time(13),
                      datetime.time(22, 58, 59), datetime.time(7, 13, 43)],
            "timestamp_": [pd.Timestamp("2016"), pd.Timestamp("2017"),
                           pd.Timestamp('2017-11-28 23:55:59.342380',
                           tz='US/Eastern'),
                           pd.Timestamp('2018-11-28 23:55:59.342380',
                           tz='Asia/Calcutta')],
            "date_": [datetime.date(2016, 1, 1), datetime.date(2017, 1, 1),
                      datetime.date(2017, 11, 28),
                      datetime.date(2018, 11, 28)],
        }, columns=['boolean_', 'smallint_', 'int_', 'bigint_', 'float_',
                    'double_', 'varchar_', 'text_', 'time_', 'timestamp_',
                    'date_'])
        con.load_table_columnar("all_types", data, preserve_index=False)

        result = list(c.execute("select * from all_types"))
        expected = [(1, 0, 0, 0, 0.0, 0.0, 'a', 'a',
                    datetime.time(0, 11, 59),
                    datetime.datetime(2016, 1, 1, 0, 0),
                    datetime.date(2016, 1, 1)),
                    (0, 1, 1, 1, 1.0, 1.0, 'b', 'b',
                    datetime.time(13, 0),
                    datetime.datetime(2017, 1, 1, 0, 0),
                    datetime.date(2017, 1, 1)),
                    (1, 0, 0, 0, 0.0, 0.0, 'a', 'a',
                    datetime.time(22, 58, 59),
                    datetime.datetime(2017, 11, 29, 4, 55, 59),
                    datetime.date(2017, 11, 28)),
                    (0, 1, 1, 1, 1.0, 1.0, 'b', 'b',
                    datetime.time(7, 13, 43),
                    datetime.datetime(2018, 11, 28, 18, 25, 59),
                    datetime.date(2018, 11, 28))]

        assert result == expected
        c.execute('drop table if exists all_types;')

    def test_load_table_columnar_arrow_all(self, con):

        c = con.cursor()
        c.execute('drop table if exists all_types;')
        create = textwrap.dedent('''\
        create table all_types (
            boolean_ BOOLEAN,
            smallint_ SMALLINT,
            int_ INT,
            bigint_ BIGINT,
            float_ FLOAT,
            double_ DOUBLE,
            varchar_ VARCHAR(40),
            text_ TEXT,
            time_ TIME,
            timestamp_ TIMESTAMP,
            date_ DATE
        );''')
        # skipping decimal for now
        c.execute(create)

        names = ['boolean_', 'smallint_', 'int_', 'bigint_',
                 'float_', 'double_', 'varchar_', 'text_',
                 'time_', 'timestamp_', 'date_']

        columns = [pa.array([True, False, None], type=pa.bool_()),
                   pa.array([1, 0, None]).cast(pa.int16()),
                   pa.array([1, 0, None]).cast(pa.int32()),
                   pa.array([1, 0, None]),
                   pa.array([1.0, 1.1, None]).cast(pa.float32()),
                   pa.array([1.0, 1.1, None]),
                   # no fixed-width string
                   pa.array(['a', 'b', None]),
                   pa.array(['a', 'b', None]),
                   (pa.array([1, 2, None]).cast(pa.int32())
                    .cast(pa.time32('s'))),
                   pa.array([datetime.datetime(2016, 1, 1, 12, 12, 12),
                             datetime.datetime(2017, 1, 1), None]),
                   pa.array([datetime.date(2016, 1, 1),
                             datetime.date(2017, 1, 1), None])]
        table = pa.Table.from_arrays(columns, names=names)
        con.load_table_arrow("all_types", table)
        c.execute('drop table if exists all_types;')

    def test_select_null(self, con):
        con.execute("drop table if exists pymapd_test_table;")
        con.execute("create table pymapd_test_table (a int);")
        con.execute("insert into pymapd_test_table VALUES (1);")
        con.execute("insert into pymapd_test_table VALUES (null);")
        # the test

        c = con.cursor()
        result = c.execute("select * from pymapd_test_table")
        expected = [(1,), (None,)]
        assert result.fetchall() == expected

        # cleanup
        con.execute("drop table if exists pymapd_test_table;")

    def test_create_table(self, con):

        con.execute("drop table if exists test_create_table;")
        df = pd.DataFrame({"A": [1, 2], "B": [1., 2.]})
        con.create_table("test_create_table", df)
        con.execute("drop table if exists test_create_table;")

    def test_load_table_creates(self, con):

        data = pd.DataFrame({
            "boolean_": [True, False],
            "smallint_cast": np.array([0, 1], dtype=np.int8),
            "smallint_": np.array([0, 1], dtype=np.int16),
            "int_": np.array([0, 1], dtype=np.int32),
            "bigint_": np.array([0, 1], dtype=np.int64),
            "float_": np.array([0, 1], dtype=np.float32),
            "double_": np.array([0, 1], dtype=np.float64),
            "varchar_": ["a", "b"],
            "text_": ['a', 'b'],
            "time_": [datetime.time(0, 11, 59), datetime.time(13)],
            "timestamp_": [pd.Timestamp("2016"), pd.Timestamp("2017")],
            "date_": [datetime.date(2016, 1, 1), datetime.date(2017, 1, 1)],
        }, columns=['boolean_', 'smallint_', 'int_', 'bigint_', 'float_',
                    'double_', 'varchar_', 'text_', 'time_', 'timestamp_',
                    'date_'])

        con.execute("drop table if exists test_load_table_creates;")
        con.load_table("test_load_table_creates", data, create=True)
        con.execute("drop table if exists test_load_table_creates;")

    def test_array_in_result_set(self, con):

        # text
        con.execute("DROP TABLE IF EXISTS test_lists;")
        con.execute("CREATE TABLE IF NOT EXISTS test_lists \
                    (col1 TEXT, col2 TEXT[]);")

        row = [("row1", "{hello,goodbye,aloha}"),
               ("row2", "{hello2,goodbye2,aloha2}")]

        con.load_table_rowwise("test_lists", row)
        ans = con.execute("select * from test_lists").fetchall()

        expected = [('row1', ['hello', 'goodbye', 'aloha']),
                    ('row2', ['hello2', 'goodbye2', 'aloha2'])]

        assert ans == expected

        # int
        con.execute("DROP TABLE IF EXISTS test_lists;")
        con.execute("CREATE TABLE IF NOT EXISTS test_lists \
                    (col1 TEXT, col2 INT[]);")

        row = [("row1", "{10,20,30}"), ("row2", "{40,50,60}")]

        con.load_table_rowwise("test_lists", row)
        ans = con.execute("select * from test_lists").fetchall()

        expected = [('row1', [10, 20, 30]), ('row2', [40, 50, 60])]

        assert ans == expected

        # timestamp
        con.execute("DROP TABLE IF EXISTS test_lists;")
        con.execute("CREATE TABLE IF NOT EXISTS test_lists \
                    (col1 TEXT, col2 TIMESTAMP[]);")

        row = [("row1",
                "{2019-03-02 00:00:00,2019-03-02 00:00:00,2019-03-02 00:00:00}"),  # noqa
               ("row2",
                "{2019-03-02 00:00:00,2019-03-02 00:00:00,2019-03-02 00:00:00}")]  # noqa

        con.load_table_rowwise("test_lists", row)
        ans = con.execute("select * from test_lists").fetchall()

        expected = [('row1',
                    [datetime.datetime(2019, 3, 2, 0, 0),
                     datetime.datetime(2019, 3, 2, 0, 0),
                     datetime.datetime(2019, 3, 2, 0, 0)]),
                    ('row2',
                    [datetime.datetime(2019, 3, 2, 0, 0),
                     datetime.datetime(2019, 3, 2, 0, 0),
                     datetime.datetime(2019, 3, 2, 0, 0)])]

        assert ans == expected

        # date
        con.execute("DROP TABLE IF EXISTS test_lists;")
        con.execute("CREATE TABLE IF NOT EXISTS test_lists \
                    (col1 TEXT, col2 DATE[]);")

        row = [("row1", "{2019-03-02,2019-03-02,2019-03-02}"),
               ("row2", "{2019-03-02,2019-03-02,2019-03-02}")]

        con.load_table_rowwise("test_lists", row)
        ans = con.execute("select * from test_lists").fetchall()

        expected = [('row1',
                     [datetime.date(2019, 3, 2),
                      datetime.date(2019, 3, 2),
                      datetime.date(2019, 3, 2)]),
                    ('row2',
                     [datetime.date(2019, 3, 2),
                      datetime.date(2019, 3, 2),
                      datetime.date(2019, 3, 2)])]

        assert ans == expected

        # time
        con.execute("DROP TABLE IF EXISTS test_lists;")
        con.execute("CREATE TABLE IF NOT EXISTS test_lists \
                    (col1 TEXT, col2 TIME[]);")

        row = [("row1", "{23:59:00,23:59:00,23:59:00}"),
               ("row2", "{23:59:00,23:59:00,23:59:00}")]

        con.load_table_rowwise("test_lists", row)
        ans = con.execute("select * from test_lists").fetchall()

        expected = [('row1',
                     [datetime.time(23, 59), datetime.time(23, 59),
                      datetime.time(23, 59)]),
                    ('row2',
                     [datetime.time(23, 59), datetime.time(23, 59),
                      datetime.time(23, 59)])]

        assert ans == expected
        con.execute("DROP TABLE IF EXISTS test_lists;")

    def test_upload_pandas_categorical(self, con):

        con.execute("DROP TABLE IF EXISTS test_categorical;")

        df = pd.DataFrame({"A": ["a", "b", "c", "a"]})
        df["B"] = df["A"].astype('category')

        # test that table created correctly when it doesn't exist on server
        con.load_table("test_categorical", df)
        ans = con.execute("select * from test_categorical").fetchall()

        assert ans == [('a', 'a'), ('b', 'b'), ('c', 'c'), ('a', 'a')]

        assert con.get_table_details("test_categorical") == \
            [ColumnDetails(name='A', type='STR', nullable=True, precision=0,
                           scale=0, comp_param=32, encoding='DICT',
                           is_array=False),
             ColumnDetails(name='B', type='STR', nullable=True, precision=0,
                           scale=0, comp_param=32, encoding='DICT',
                           is_array=False)]

        # load row-wise
        con.load_table("test_categorical", df, method="rows")

        # load columnar
        con.load_table("test_categorical", df, method="columnar")

        # load arrow
        con.load_table("test_categorical", df, method="arrow")

        # test end result
        df_ipc = con.select_ipc("select * from test_categorical")
        assert df_ipc.shape == (16, 2)

        res = df.append([df, df, df]).reset_index(drop=True)
        res["A"] = res["A"].astype('category')
        res["B"] = res["B"].astype('category')
        assert pd.DataFrame.equals(df_ipc, res)

        # test that input df wasn't mutated
        # original input is object, categorical
        # to load via Arrow, converted internally to object, object
        assert is_object_dtype(df["A"])
        assert is_categorical_dtype(df["B"])
        con.execute("DROP TABLE IF EXISTS test_categorical;")

    def test_insert_unicode(self, con):

        """INSERT Unicode using bind_params"""

        c = con.cursor()
        c.execute('drop table if exists text_holder;')
        create = ('create table text_holder (the_text text);')
        c.execute(create)
        first = {"value": "我和我的姐姐吃米饭鸡肉"}
        second = {"value": "El camina a case en bicicleta es relajante"}

        i1 = "INSERT INTO text_holder VALUES ( :value );"

        c.execute(i1, parameters=first)
        c.execute(i1, parameters=second)

        c.execute('drop table if exists text_holder;')

    def test_execute_leading_space_and_params(self, con):

        # https://github.com/omnisci/pymapd/issues/263

        """Ensure that leading/trailing spaces in execute statements
           don't cause issues
        """

        c = con.cursor()
        c.execute('drop table if exists test_leading_spaces;')
        create = ('create table test_leading_spaces (the_text text);')
        c.execute(create)
        first = {"value": "我和我的姐姐吃米饭鸡肉"}
        second = {"value": "El camina a case en bicicleta es relajante"}

        i1 = """

                    INSERT INTO test_leading_spaces


                    VALUES ( :value );

                            """

        c.execute(i1, parameters=first)
        c.execute(i1, parameters=second)

        c.execute('drop table if exists test_leading_spaces;')



from pymapd.cursor import make_row_results_set
from omnisci.mapd.ttypes import (TRowSet, TColumnType, TColumn, TColumnData,
                                 TQueryResult)
from omnisci.common.ttypes import TTypeInfo


class TestRowResults:

    def test_nulls_handled(self):

        rs = TQueryResult(TRowSet(
            row_desc=[
                TColumnType(col_name='a',
                            col_type=TTypeInfo(type=0, nullable=True)),
                TColumnType(col_name='b',
                            col_type=TTypeInfo(type=1, nullable=True)),
                TColumnType(col_name='c',
                            col_type=TTypeInfo(type=2, nullable=True)),
                TColumnType(col_name='d',
                            col_type=TTypeInfo(type=3, nullable=True)),
                TColumnType(col_name='e',
                            col_type=TTypeInfo(type=4, nullable=True)),
                TColumnType(col_name='f',
                            col_type=TTypeInfo(type=5, nullable=True)),
                TColumnType(col_name='g',
                            col_type=TTypeInfo(type=6, nullable=True)),
                TColumnType(col_name='h',
                            col_type=TTypeInfo(type=7, nullable=True)),
                TColumnType(col_name='i',
                            col_type=TTypeInfo(type=8, nullable=True)),
                TColumnType(col_name='j',
                            col_type=TTypeInfo(type=9, nullable=True)),
                TColumnType(col_name='k',
                            col_type=TTypeInfo(type=10, nullable=True)),
            ],
            rows=[],
            columns=[
                TColumn(data=TColumnData(int_col=[-2147483648]), nulls=[True]),
                TColumn(data=TColumnData(int_col=[-2147483648]), nulls=[True]),
                TColumn(data=TColumnData(int_col=[-2147483648]), nulls=[True]),
                TColumn(data=TColumnData(real_col=[-2147483648]), nulls=[True]),  # noqa
                TColumn(data=TColumnData(real_col=[-2147483648]), nulls=[True]),  # noqa
                TColumn(data=TColumnData(real_col=[-2147483648]), nulls=[True]),  # noqa
                TColumn(data=TColumnData(str_col=[-2147483648]), nulls=[True]),
                TColumn(data=TColumnData(int_col=[-2147483648]), nulls=[True]),
                TColumn(data=TColumnData(int_col=[-2147483648]), nulls=[True]),
                TColumn(data=TColumnData(int_col=[-2147483648]), nulls=[True]),
                TColumn(data=TColumnData(int_col=[-2147483648]), nulls=[True]),
            ],
            is_columnar=True))

        result = list(make_row_results_set(rs))
        assert result == [(None,) * 11]
